/*

*/

import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.locks.Condition;

/**
   This simulates a first-in-first-out (stack) buffer.
*/
public class Buffer_bounded extends Buffer
{
   private final ReentrantLock lock;
   private final Condition emptyCondition;
   private boolean empty;
   private final Condition fullCondition;
   private boolean full;

   /**
      Set a default buffer size of 80.
   */
   Buffer_bounded(final int currentIndex)
   {
      this(currentIndex, 80);
   }

   /**
      Create a buffer with a mutual exclusion lock,
      a condition variable that prevents buffer
      underflow, and another condition variable
      that prevents buffer overflow.
   */
   Buffer_bounded(final int currentIndex, final int SIZE)
   {
      super(currentIndex, SIZE); //Initialize a partially filled buffer.

      this.lock = new ReentrantLock();
      this.fullCondition  = lock.newCondition();
      this.emptyCondition = lock.newCondition();
      this.empty = (currentIndex == 0);
      this.full  = (currentIndex == SIZE);
   }

   @Override
   public void put(char c)
   {
      try
      {
         this.lock.lock(); // mutual exclusion
         try
         {
            while (this.full)
            {
               this.fullCondition.await();
            }
            ++currentIndex;
            Thread.yield(); // force race conditions to happen more often
            buf[currentIndex] = c;
            this.empty = (currentIndex == 0);
            this.full = (currentIndex == SIZE);
            this.emptyCondition.signal(); // can use signalAll OR signal()
         }
         finally
         {
            this.lock.unlock();
         }
      }
      catch (InterruptedException e)
      {
         // Without this line, the program can deadlock.
         Thread.currentThread().interrupt();
      }
   }


   @Override
   public void get(char c)
   {
      try
      {
         this.lock.lock(); // mutual exclusion
         try
         {
            while (this.empty)
            {
               this.emptyCondition.await();
            }
            buf[currentIndex] = c;
            Thread.yield(); // force race conditions to happen more often
            --currentIndex;
            this.empty = (currentIndex == 0);
            this.full = (currentIndex == SIZE);
            this.fullCondition.signal(); // can use signalAll OR signal()
         }
         finally
         {
            this.lock.unlock();
         }
      }
      catch (InterruptedException e)
      {
         // Without this line, the program can deadlock.
         Thread.currentThread().interrupt();
      }
   }
}
