/*

*/

import java.util.concurrent.locks.ReentrantLock;
import java.util.concurrent.locks.Condition;

/**
   This simulates a first-in-first-out (stack) buffer.
*/
public class Buffer_unbounded extends Buffer
{
   private final ReentrantLock lock;
   private final Condition emptyCondition;
   private boolean empty;

   /**
      Set a default buffer size of 80.
   */
   Buffer_unbounded(final int currentIndex)
   {
      this(currentIndex, 80);
   }

   /**
      Create a buffer with a mutual exclusion lock
      and a condition variable that prevents buffer
      underflow.
   */
   Buffer_unbounded(final int currentIndex, final int SIZE)
   {
      super(currentIndex, SIZE); // Initialize a partially filled buffer.

      this.lock = new ReentrantLock();
      this.emptyCondition = lock.newCondition();
      this.empty = (currentIndex == 0);
   }


   @Override
   public void put(char c)
   {
      this.lock.lock(); // mutual exclusion
      try
      {
         ++currentIndex;
         Thread.yield(); // force race conditions to happen more often
         buf[currentIndex] = c;
         this.empty = (currentIndex == 0);
         this.emptyCondition.signal(); // can use signalAll OR signal()
      }
      finally
      {
         this.lock.unlock();
      }
   }


   @Override
   public void get(char c)
   {
      try
      {
         this.lock.lock(); // mutual exclusion
         try
         {
            while (this.empty)
            {
               this.emptyCondition.await();
            }
            buf[currentIndex] = c;
            Thread.yield(); // force race conditions to happen more often
            --currentIndex;
            this.empty = (currentIndex == 0);
         }
         finally
         {
            this.lock.unlock();
         }
      }
      catch (InterruptedException e)
      {
         // Without this line, the program can deadlock.
         Thread.currentThread().interrupt();
      }
   }
}
